﻿using Myriadbits.NTFS;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Myriadbits.DiskSpaceAnalyzer
{
	public class TopFileList : ListView
	{
		private List<RootFolder>	m_disks = new List<RootFolder>();
		private Color				m_textColorZeroSize = MainForm.m_textColorZeroSize;
		private Color				m_textColor = MainForm.m_textColor;
		private TreeViewMode		m_treeViewMode = TreeViewMode.AllFiles;
		private UnitTypes			m_currentUnit = UnitTypes.Auto;
		
		public int SortColumn { get; set; }
		public bool SortAscending { get; set; }
		public int NumberOfFilesPerDisk { get; set; }

		/// <summary>
		/// Constructor
		/// </summary>
		public TopFileList()
		{
			this.SortColumn = 0;
			this.SortAscending = true;			
			this.ColumnClick += this.onColumnClicked;
		}


		/// <summary>
		/// Initialize this class
		/// </summary>
		/// <param name="lvwTopList"></param>
		public void Initialize(int numberOfFiles = 10)
		{
			if (this.Parent != null)
				this.Parent.Text = string.Format("Top {0} files", numberOfFiles);
			this.NumberOfFilesPerDisk = numberOfFiles;
		}


		/// <summary>
		/// Change the current view mode
		/// </summary>
		/// <param name="newViewMode"></param>
		public void ChangeViewMode(TreeViewMode newViewMode)
		{
			if (newViewMode != this.m_treeViewMode)
			{
				this.m_treeViewMode = newViewMode;
				RefreshFileList();
			}
		}


		/// <summary>
		/// The current unit type has changed (bytes to MB or GB, or...)
		/// </summary>
		/// <param name="newUnitType"></param>
		public void ChangeUnit(UnitTypes newUnitType)
		{
			if (newUnitType != m_currentUnit)
			{
				m_currentUnit = newUnitType;
				RefreshFileList();
			}
		}


		/// <summary>
		/// Header click -> sort the list
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void onColumnClicked(object sender, ColumnClickEventArgs e)
		{
			if (this.SortColumn == e.Column)
			{
				this.SortAscending = !this.SortAscending;
			}
			else
			{
				this.SortAscending = true;
				this.SortColumn = e.Column;
			}

			// Refresh/update the list
			RefreshFileList();
		}


		/// <summary>
		/// Add a new tree to the list of tress
		/// </summary>
		/// <param name="treeBuilder"></param>
		public void AddRoot(RootFolder rootFolder)
		{
			m_disks.Add(rootFolder);
		}


		/// <summary>
		/// Clear the disk list
		/// </summary>
		public void ClearList()
		{
			m_disks.Clear();
		}


		/// <summary>
		/// Fill the list with data
		/// </summary>
		public void RefreshFileList()
		{
			// Clear whole list
			this.Items.Clear();

			this.m_disks = this.m_disks.OrderBy(a => a.Name).ToList();
			foreach(RootFolder root in this.m_disks)
			{
				ListViewGroup lvwGroup = new ListViewGroup(root.DriveDisplayName);
				this.Groups.Add(lvwGroup);


				List<File> topFiles = root.Parser.LargestFiles.Take(NumberOfFilesPerDisk).Select(a => a.Value).ToList();
				List<File> topFilesOutput = new List<File>(); 
				switch (this.SortColumn)
				{
					case 0: // Sort by ranking
						if (!this.SortAscending)
							topFiles.Reverse();
						// Else: Do nothing, already in correct order
						break;
					case 1: // File names
						if (this.SortAscending)
							topFiles = topFiles.OrderBy(a => a.Name).ToList();
						else
							topFiles = topFiles.OrderByDescending(a => a.Name).ToList();
						break;
					case 2: // Size of files
						if (this.SortAscending)
							topFiles = topFiles.OrderByDescending(a => a.Size).ToList();
						else
							topFiles = topFiles.OrderBy(a => a.Size).ToList();
						break;
					case 3: // Path
						if (this.SortAscending)
							topFiles = topFiles.OrderBy(a => a.Folder.FullPath).ToList();
						else
							topFiles = topFiles.OrderByDescending(a => a.Folder.FullPath).ToList();
						break;
				}

				foreach (File file in topFiles)
				{
					ListViewItem lvi = new ListViewItem((file.Rank + 1).ToString());
					lvi.SubItems.Add(file.Name);
					lvi.SubItems.Add(this.m_currentUnit.GetUnitText(file.Size));
					lvi.SubItems.Add(file.Folder.FullPath);
					lvi.Group = lvwGroup;
					lvi.Tag = file;

					if (!TreeBuilder.IsFileVisible(file, this.m_treeViewMode))
						lvi.ForeColor = m_textColorZeroSize;
					else
						lvi.ForeColor = m_textColor;

					this.Items.Add(lvi);				
				}
			}
		}


		/// <summary>
		/// Get the current selected folder
		/// </summary>
		/// <returns></returns>
		public Folder SelectedFolder
		{
			get
			{
				if (this.SelectedItems.Count > 0)
				{
					File file = this.SelectedItems[0].Tag as File;
					if (file != null)
						return file.Folder;
				}
				return null;
			}
		}


		/// <summary>
		/// Get the current selected file
		/// </summary>
		/// <returns></returns>
		public File SelectedFile
		{
			get
			{
				if (this.SelectedItems.Count > 0)
					return this.SelectedItems[0].Tag as File;
				return null;
			}
		}


	}
}
