﻿//
// NTFS - Myriadbits .NET NTFS library. 
// Fast MFT reader.
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;


namespace Myriadbits.NTFS
{
	public class NTFSAttributeReference : DataBufferReader
	{
		public NTFSAttributeReference(DataBuffer buffer, int offset)
			: base(buffer, 0, offset)
		{
		}

		public NTFSAttributeType AttributeType { get { return (NTFSAttributeType)this.GetUInt(0x00); } }
		public ushort RecordLength { get { return (this.GetUShort(0x04)); } }
		public byte AttributeNameLength { get { return (this.GetByte(0x06)); } }
		public byte OffsetToName { get { return (this.GetByte(0x07)); } }

		public ulong StartingVCN { get { return (this.GetULong(0x08)); } }
		public ulong BaseFileReference { get { return (this.GetULong(0x10)) & NTFS.referenceMask; } }
		public ushort BaseFileReferenceSequenceNumber { get { return (ushort) ((this.GetULong(0x10) >> 48) & 0xFFFF); } }
		public ushort AttributeId { get { return (this.GetUShort(0x18)); } }
		public string AttributeName { get { return this.GetString(0x1A, this.AttributeNameLength * 2); } }

		public override string ToString()
		{
			return string.Format("${0}: {1}", this.AttributeType.ToString(), this.AttributeName.ToString());
		}
	}

	public class NTFSAttributeAttributeList : NTFSAttribute
	{
		private List<NTFSAttributeReference> _attributereferences = null;
		
		public NTFSAttributeAttributeList(DataBuffer buffer, int offset) 
			: base(buffer, offset)
		{
			this._attributereferences = null;
		}

		/// <summary>
		/// Return a list of all attribute references
		/// </summary>
		public List<NTFSAttributeReference> AttributeReferences
		{
			get
			{
				if (this._attributereferences == null)
					LoadAttributeReferences();
				return this._attributereferences;
			}
		}


		private void LoadAttributeReferences()
		{
			// Get all attributes
			NTFSAttributeReference attributeReference = null;
			this._attributereferences = new List<NTFSAttributeReference>();
			int currentOffset = this._dataoffset;
			while (currentOffset < this._dataoffset + this.Length)
			{
				attributeReference = new NTFSAttributeReference(this._buffer, currentOffset);
				if (attributeReference != null)
				{
					this._attributereferences.Add(attributeReference);
					currentOffset += attributeReference.RecordLength;
				}
			}
		}


		/// <summary>
		/// Return a list of all filename attribute references
		/// </summary>
		public List<NTFSAttributeReference> AttributeFilenameReferences
		{
			get
			{
				if (this._attributereferences == null)
					LoadAttributeReferences();
				return this._attributereferences.Where(a => a.AttributeType == NTFSAttributeType.File_Name && a.StartingVCN == 0).ToList();
			}
		}


		public override string ToString()
		{
			string retVal = string.Format("$ATTRIBUTE_LIST[{0}]", this._attributereferences.Count);
			foreach (NTFSAttributeReference attributeRef in this._attributereferences)
				retVal += attributeRef.ToString() + ", ";
			return retVal;
		}
	}
}
