﻿//
// NTFS - Myriadbits .NET NTFS library. 
// Fast MFT reader.
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Threading;
using System.Collections;

//
// TODO: C:\Program Files\Windows Sidebar\Gadgets\Calendar.Gadget\images>

namespace Myriadbits.NTFS
{
	/// <summary>
	/// Small class to notify the progress to the UI
	/// </summary>
	public class NTFSParserProgress
	{
		public ulong Current { get; set; }
		public ulong Maximum { get; set; }
	}

	// A delegate type for hooking up the progress changed notifications.
	public delegate void ProgressChangedEventHandler(object sender, NTFSParserProgress progress);

	// A delegate type for hooking up the finished notification.
	public delegate void BuildTreeFinishedEventHandler(object sender, Folder mainFolder);

	// A delegate type for hooking up the logging.
	public delegate void LogHandler(object sender, string message);


	/// <summary>
	/// Parse a volume
	/// </summary>
	public class NTFSReader
	{
		private FolderCollection allFolders;

		// Events
		public event ProgressChangedEventHandler ProgressChanged;
		public event BuildTreeFinishedEventHandler BuildTreeFinished;
		public event LogHandler Logger;
		public List<KeyValuePair<ulong, File>> m_largestFiles = null;
		private NTFSVolume m_volume;

		/// <summary>
		/// Return the volume this parser is currently working on
		/// </summary>
		public NTFSVolume Volume { get { return this.m_volume; } }


		private const int MAX_SI_DICT_SIZE = 4096; // Maximum size of standard information dictionary

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="listLog"></param>
		public NTFSReader()
		{
			this.ReadStandardInformation = false;
		}

		// TODO Implement IDispose
		// call volume.Close();


		/// <summary>
		/// Generic properties
		/// </summary>
		public bool IsNTFS { get; private set; }

		/// <summary>
		/// Statistical information
		/// </summary>
		public ulong ElapsedMilliseconds { get; set; }
		public ulong TotalNumberOfFiles { get; set; }
		public ulong TotalNumberOfDirectories { get; set; }
		public ulong TotalNumberOfBytes { get; set; }

		/// <summary>
		/// Properties of the request
		/// </summary>
		private bool ShowHiddenFolders { get; set; }
		private bool ShowHiddenFiles { get; set; }
		private int MaximumOrderedFileListSize { get; set; }
		public string Drive { get; set; }
		public bool ReadStandardInformation { get; set; }


		/// <summary>
		/// Return the root folder of the drive
		/// </summary>
		public Folder RootFolder 
		{ 
			get
			{
				if (this.allFolders != null)
					return this.allFolders.RootFolder;
				return null;
			}
		}
		
	
		/// <summary>
		/// Scan an entire NTFS volume, build the tree with sizes
		/// The scanning is performed in the background, use the events to be notified of the progress and logging
		/// </summary>
		/// <param name="drive"></param>
		public void ReadVolumeTree(string drive, int maxOrderedFileListSize, bool showHiddenFolders = true, bool showHiddenFiles = false)
		{	
			// Store properties
			string driveName = drive.Substring(0, 1);
			this.Drive = driveName;
			this.MaximumOrderedFileListSize = maxOrderedFileListSize;
			this.ShowHiddenFolders = showHiddenFolders;
			this.ShowHiddenFiles = showHiddenFiles;

			ThreadPool.QueueUserWorkItem(this.ThreadPoolCallback, this);
		}

		
		/// <summary>
		/// Background process
		/// </summary>
		/// <param name="threadContext"></param>
	    public void ThreadPoolCallback(Object threadContext)
		{
			try
			{
				// Read NTFS header
				UnmanagedFileLoader fl = new UnmanagedFileLoader(string.Format("\\\\.\\{0}:", this.Drive));
				byte[] header = fl.ReadBytes(1024);

				// Is this drive a NTFS drive???
				Stopwatch sw = Stopwatch.StartNew();
				this.IsNTFS = (header[3] == 'N' && header[4] == 'T' && header[5] == 'F' && header[6] == 'S');
				if (this.IsNTFS)
				{
					// Yes, NTFS
					Log("NTFS");

					// Read general NTFS data
					NTFSHeader ntfsHeader = new NTFSHeader(new DataBuffer(header));

					Log("Driver {0}", this.Drive);
					Log("- SectorSize: {0}, SectorsPerCluster: {1}, MFT: {2}", ntfsHeader.BytesPerSect, ntfsHeader.SectorsPerCluster, ntfsHeader.MFT);


					// Reserve some capacity for the folders
					this.allFolders = new FolderCollection();

					// Counters
					ulong fileCounter = 0;
					ulong dirCounter = 0;
					ulong refCounter = 0;
					ulong additionalReads = 0;

					// Determine the offset of the MFT (Master File Table)
					ulong blockSize = ntfsHeader.BlockSize;
					ulong offsetRead = ntfsHeader.MFTOffset;

					Log("- BlockSize: {0}", blockSize);

					// Read part of the MFT
					fl.Seek(offsetRead);

					// Read a part of the MFT
					byte[] data = new byte[blockSize];
					uint reads = fl.ReadFile(ref data, (int)blockSize);
					DataBuffer mftdata = new DataBuffer(data);
					NTFSFileRecord mftFile = new NTFSFileRecord(mftdata, (int) ntfsHeader.BytesPerSect);

					// Create the dataruns				
					NTFSAttribute dataAttribute = mftFile.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Data).FirstOrDefault();
					NTFSDataRuns dataRuns = new NTFSDataRuns(dataAttribute);

					ulong totalRead = 0;

					// Notify the progress
					NTFSParserProgress progress = new NTFSParserProgress();
					progress.Maximum = dataRuns.TotalSize * ntfsHeader.BytesPerCluster;
					progress.Current = 0;
					ProgressChanged(this, progress);
					//form.progressBar1.Maximum = (int) (dataRuns.TotalSize * SectperCluster * BytesPerSect);

					// Calculate permission masks
					int folderPermissionMask = 0x0040;
					int filePermissionMask = 0x0040;
					if (!this.ShowHiddenFolders) folderPermissionMask |= 0x0002;
					if (!this.ShowHiddenFiles) filePermissionMask |= 0x0006;

					// Loop through the dataruns
					int dataRunCounter = 0;
					int validCounter = 0;
					int seekCounter = 0;
					Stopwatch seekStopwatch = new Stopwatch();

					this.m_volume = new NTFSVolume();
					if (!this.m_volume.Open(this.Drive))
					{
						Log("FAILED to open the volume");
					}

					List< KeyValuePair<ulong, NTFSAttributeStandardInformation> > dicBaseSIs = new List< KeyValuePair< ulong, NTFSAttributeStandardInformation> > (MAX_SI_DICT_SIZE);


					foreach (NTFSDataRun dataRun in dataRuns.DataRuns)
					{
						Log("  - Datarun {0}, Length: {1}, Offset: {2}", dataRunCounter, dataRun.Length, dataRun.Position);

						// Find the correct position
						offsetRead = dataRun.Position * ntfsHeader.BytesPerCluster;
						ulong offsetReadEnd = (dataRun.Length + dataRun.Position) * ntfsHeader.BytesPerCluster;

						bool fQuit = false;
						while (offsetRead < offsetReadEnd)
						{
							// Read part of the MFT
							fl.Seek(offsetRead);

							// Read a large block at once
							ulong readSize = blockSize;
							if (offsetRead + readSize > offsetReadEnd)
								readSize = offsetReadEnd - offsetRead;
							reads = fl.ReadFile(ref data, (int)readSize);

							// Create a buffer around the data
							mftdata = new DataBuffer(data);
							long offsetMemory = 0;
							NTFSFileRecord filerecord = null;
							NTFSFileRecord lastFile = null;
							NTFSFileRecord filerecordBase = null;
							int errorCounter = 0;
							while (offsetMemory < mftdata.Length && (ulong)offsetMemory + offsetRead < offsetReadEnd)
							{
								lastFile = filerecord;
								filerecord = new NTFSFileRecord(mftdata, (int) ntfsHeader.BytesPerSect, (int)offsetMemory);
								int nextRecordJump = filerecord.AllocatedSizeOfRecord;
								if (!filerecord.IsValid)
								{
									Log("  - Error, Invalid filerecord (found {0} valids before)!!", validCounter);
									if (lastFile != null)
										Log("*Last:    " + lastFile.ToString());
									Log("*Current: " + filerecord.ToString());
									errorCounter++;
									if (errorCounter > 100)
									{
										fQuit = true; // Emergency exit
										break;
									}
									nextRecordJump = 1024;
								}
								validCounter++;

								if (filerecord.IsInUse)
								{									
									// First name attribute is the actual name, the others are symbolic links to the same file...
									// But all folder sizes in windows actually include the hardlink as well! (buggy)

									filerecordBase = null;
									NTFSAttributeStandardInformation standardInformation = null;
									if (filerecord.BaseFileRecordReference == 0)
									{
										standardInformation = filerecord.StandardAttributeInformation;
									}
									else
									{
										// Note: storing this info in a temporary array/list has a negative impact on performance, 
										// search takes longer then disk retrieval... (is this correct?)

										// To get the dos file attributes, the standard information attirbute is required
										// for this record, this is stored in the base file, we can retrieve that but that
										// takes up a huge amount of extra performance
										if (this.ReadStandardInformation)
										{
											seekStopwatch.Start();
											filerecordBase = this.m_volume.ReadFileRecord(filerecord.BaseFileRecordReference);
											standardInformation = filerecordBase.StandardAttributeInformation;
											seekStopwatch.Stop();
										}
										else
										{
											// Do not use the standard information => much faster
											standardInformation = null;
										}
										seekCounter++;
									}


									// Add the folder or file
									if (filerecord.IsDirectory)
									{
										bool addFolder = true;
										if (this.ReadStandardInformation && (standardInformation.DOSFilePermissions & folderPermissionMask) != 0)
											addFolder = false;

										// Not required: AddIndexTableFiles(filerecord);

										// Add this directory to the list of directories
										if (addFolder)
										{
											if (this.allFolders.AddFolder(filerecord, standardInformation, folderPermissionMask))
												dirCounter++;
										}
									}
									else
									{
										bool addFile = true;
										if (this.ReadStandardInformation && (standardInformation.DOSFilePermissions & filePermissionMask) != 0)
											addFile = false;

										// Add this file to a specific folder
										if (addFile)
										{
											if (AddFile(filerecord, standardInformation, filePermissionMask, this.m_volume))
												fileCounter++;
										}
									}
								}
								offsetMemory += filerecord.AllocatedSizeOfRecord; // next record
								refCounter++;
							}
							totalRead += readSize;

							// Notify the progress
							progress.Current = totalRead;
							if (refCounter % 10 == 0) // Do not update to much
								ProgressChanged(this, progress);

							if (fQuit) break; // Emergency exit

							offsetRead += blockSize;
						}
						dataRunCounter++;
					}

					// Set the local properties
					this.ElapsedMilliseconds = (ulong)sw.ElapsedMilliseconds;

					// Build the tree
					sw.Restart();
					this.allFolders.FinalizeTree(this);
					Log("- Tree build in {0} ms", sw.ElapsedMilliseconds);

					// Sort the files
					sw.Restart();
					CalculateLargestFiles();
					Log("- Files sorted in {0} ms", sw.ElapsedMilliseconds);

					this.TotalNumberOfFiles = fileCounter;
					this.TotalNumberOfDirectories = dirCounter;
					if (RootFolder != null)
						this.TotalNumberOfBytes = RootFolder.Size;

					Log("- Processtime {0} ms", this.ElapsedMilliseconds);
					Log("- {0} File(s) {1} bytes", fileCounter, RootFolder.Size);
					Log("- {0} Dir(s)", dirCounter);
					Log("- {0} Indirect data attributes found", additionalReads);
					Log("- {0} Extra seeks required (took {1} ms)", seekCounter, seekStopwatch.ElapsedMilliseconds);
					Log("- MFT size: {0} Mb", (int)(dataRuns.TotalSize * ntfsHeader.SectorsPerCluster) / (1024 * 1024));
					Log("");

					if (this.m_largestFiles.Count() == 0)
						Log("- There are NO files on this volume!");
					else
					{
						Log("- 10 Largest files:");
						for (int n = 0; n < 10 && n < this.m_largestFiles.Count(); n++)
							Log("   - {0} - File: {1}, Size: {2}", n, this.m_largestFiles[n].Key, this.m_largestFiles[n].Value);
					}
					
					// Fire the progress changed event
					progress.Current = progress.Maximum;
					ProgressChanged(this, progress);

					// Fire the finished event
					BuildTreeFinished(this, this.RootFolder);
				}
				else
				{
					Log("Sorry, the selected drive is NOT an NTFS drive");

					// Fire the finished event
					BuildTreeFinished(this, null);
				}
			}
			catch(Exception ex)
			{
				Log("ERROR during collecting! {0}", ex.Message);
				BuildTreeFinished(this, null);
			}
		}


		/// <summary>
		/// The filerecord is a file, add the file to the parent folder
		/// </summary>
		/// <returns></returns>
		public bool AddFile(NTFSFileRecord filerecord, NTFSAttributeStandardInformation standardAttr, int filePermissionMask, NTFSVolume volume)
		{
			bool retVal = false;
			//if (standardAttr != null)
			{
				// 
				// FILE
				//
				// Do not add masked out files
				ulong referenceNumber = filerecord.ReferenceNumber;
				if (filerecord.BaseFileRecordReference != 0)
					referenceNumber = filerecord.BaseFileRecordReference;


				// Same file reference but different namespace, only add 1
				// Same file reference but same namespace, add each one
				if (referenceNumber > 34)
				{
					//
					// Loop through all name attributes
					ulong dataSize = filerecord.GetFilesize(volume);
					for (byte nameIndex = 0; nameIndex < filerecord.NameAttributes.Count; nameIndex++)
					{
						NTFSAttributeFilename nameAttr = filerecord.NameAttributes[nameIndex];

						// Create a unique reference number so we can trace the index of the nameattribute for this file
						ulong refNumber = filerecord.CreateUniqueReferenceNumber(0, nameIndex);

						// Add the file size to the folder + parent folders
						this.allFolders.AddFile(refNumber, nameAttr.ParentReference, dataSize, nameAttr.Filename);

						if (nameAttr.Filename.StartsWith("Microsoft-Windows-ImageBasedSetup-IE-Package-Base-Downlevel~31bf3856ad36"))
							retVal = true;

						retVal = true;
					}

					// And now the references (if any)
					//if (filerecord.NameAttributeReferences != null)
					//{
					//	for (byte nameRefIndex = 0; nameRefIndex < filerecord.NameAttributeReferences.Count; nameRefIndex++ )
					//	{
					//		NTFSAttributeReference nameRef = filerecord.NameAttributeReferences[nameRefIndex];
					//		NTFSFileRecord filenameRecord = volume.ReadFileRecord(nameRef.BaseFileReference); // Only interested in the file reference number
					//		if (filenameRecord != null)
					//		{
					//			NTFSAttributeFilename nameAttrRef = filenameRecord.NameAttribute;
					//			if (nameAttrRef != null)
					//			{
					//				ulong refNumber = filerecord.CreateUniqueReferenceNumber(nameRefIndex, 0);
					//				AddFileToFolder(referenceNumber, nameAttrRef.ParentReference, dataSize);
					//			}
					//		}
					//	}
					//}
				}
			}
			return retVal;
		}


		/// <summary>
		/// Returns the folder path of a specific folder
		/// </summary>
		/// <returns></returns>
		public string GetFolderPath(ulong referenceNumber)
		{
			string path = "";
			Folder fol = this.allFolders[referenceNumber];

			// Parent already exist, add the data to the parent
			while (fol != null)
			{
				if (string.IsNullOrEmpty(path))
					path = fol.Name;
				else
					path = string.Format("{0}\\{1}", fol.Name, path);
				fol = fol.ParentFolder;
			}				
			return path;
		}

		/// <summary>
		/// Background process
		/// </summary>
		/// <param name="threadContext"></param>
	    public static bool IsDriveNTFS(string drive)
		{
			try
			{
				// Use the first char
				string driveName = drive.Substring(0, 1);

				// Read NTFS header
				UnmanagedFileLoader fl = new UnmanagedFileLoader(string.Format("\\\\.\\{0}:", driveName));
				byte[] header = fl.ReadBytes(1024); 
			
				// Is this drive a NTFS drive???
				return (header[3] == 'N' && header[4] == 'T' && header[5] == 'F' && header[6] == 'S');
			}
			catch (Exception)
			{
				return false;
			}
		}


		/// <summary>
		/// Write some logging information to a listbox
		/// </summary>
		/// <param name="format"></param>
		/// <param name="args"></param>
		private void Log(string format, params object[] args) 
		{
			if (this.Logger != null)
			{
				if (this.m_volume != null)
					this.Logger(this, string.Format("[{0}] {1}", this.m_volume.Drive, string.Format(format, args)));
				else
					this.Logger(this, string.Format(format, args));
			}
		}


		/// <summary>
		/// Read a single file record
		/// </summary>
		/// <param name="requestedFileRecordReference"></param>
		/// <returns></returns>
		private NTFSFileRecord ReadFileRecord(UnmanagedFileLoader fl, NTFSDataRuns dataRuns, NTFSHeader ntfsHeader, ulong requestedFileRecordReference)
		{				
			// Loop through the dataruns
			int datarunCounter = 0;
			ulong datarunStartFileRecord = 0;
			foreach (NTFSDataRun dataRun in dataRuns.DataRuns)
			{
				ulong datarunTotalFileRecords = dataRun.Length * ntfsHeader.BytesPerCluster / 1024;
				if (requestedFileRecordReference >= datarunStartFileRecord && requestedFileRecordReference < datarunStartFileRecord + datarunTotalFileRecords)
				{
					// Requested file reference is in this datarun

					// Calculate the requested position
					ulong offsetRead = dataRun.Position * ntfsHeader.BytesPerCluster;
					offsetRead += (requestedFileRecordReference - datarunStartFileRecord) * 1024;

					// Read the filerecords
					byte[] data = new byte[1024];
					fl.Seek(offsetRead);
					uint numberOfBytes = fl.ReadFile(ref data, (int)1024);

					// Create a buffer around the data
					DataBuffer dataBuffer = new DataBuffer(data);
					NTFSFileRecord fileRecord = new NTFSFileRecord(dataBuffer, (int)ntfsHeader.BytesPerSect);

					// Return the found file record
					return fileRecord;
				}

				// Next datarun
				datarunStartFileRecord += datarunTotalFileRecords;
				datarunCounter++;
			}
			return null;
		}



		/// <summary>
		/// Add files found in index tables
		/// 
		/// NOTE:	It seems that this is not needed when scanning ALL file records, the files in the index
		///			tables can also be found in the file records.
		///			(Might be usefull later on...)
		/// </summary>
		/// <param name="filerecord"></param>
		void AddIndexTableFiles(NTFSFileRecord filerecord)
		{
			// Also add all filerecords that are part of the resident index table
			NTFSAttributeIndexRoot indexRoot = filerecord.IndexRoot;
			if (indexRoot != null && indexRoot.IndexHeader != null)
			{
				for (byte indexIndex = 0; indexIndex < indexRoot.IndexHeader.IndexRecords.Count; indexIndex++)
				{
					NTFSIndexRoot_IndexRecord indexRecord = indexRoot.IndexHeader.IndexRecords[indexIndex];
					//if (indexRecord.Filename.StartsWith("Microsoft-Windows-ImageBasedSetup-IE-Package-Base-Downlevel"))
					//	indexIndex = indexIndex;

					ulong refNumber = indexRecord.CreateUniqueReferenceNumber(indexIndex, 0);

					//// Only if the file has not yet been added to the folder
					////if (currentFolder.Files == null || (currentFolder.Files != null && !currentFolder.Files.Contains(refNumber)))
					if (indexRecord.FileReferenceNumber > filerecord.ReferenceNumber)
					{
						if (indexRecord.FileReferenceNumber > 34) // First X are reserved for system files (i.e. $MFT, $OBJ etc))
						{
							// Add the file size to the folder + parent folders
							// TODO How about dos file permissions here??
							//NTFSFileRecord baseFileRecord = volume.ReadFileRecord(indexRecord.FileReferenceNumber);
							//if (baseFileRecord != null)
							{
								// TODO How about dos file permissions here??
								//ulong dataSize = baseFileRecord.GetFilesize(volume);
								ulong dataSize = indexRecord.DataSize;

								// Add the file size to the folder + parent folders
								//AddFileToFolder(refNumber, filerecord.ReferenceNumber, dataSize);
								//fileCounter++;
							}
						}
					}
				}
			}
		}


		/// <summary>
		/// Return a list of the largest X files
		/// </summary>
		public List<KeyValuePair<ulong, File>> LargestFiles
		{
			get
			{
				return this.m_largestFiles;
			}
		}


		/// <summary>
		/// (Re)determine the largest files
		/// </summary>
		public void CalculateLargestFiles()
		{
			this.m_largestFiles = this.allFolders.FindLargest(this.MaximumOrderedFileListSize);
			// Now set the rank of each of these files
			int rankCounter = 0; // Rank starts at zero, -1 = not set
			foreach (KeyValuePair<ulong, File> kvp in this.m_largestFiles)
				kvp.Value.Rank = rankCounter++;
		}


		/// <summary>
		/// Find all files that match a certain search mask (* and ? allowed)
		/// </summary>
		/// <param name="searchMask"></param>
		/// <returns></returns>
		public List<File> FindFiles(string searchMask)
		{
			return this.allFolders.FindFiles(searchMask);
		}

	}
}
